module DepthAnalysis.Plugin.Linkables (newToken, incrementToken, dumpStatistics, thenIO) where

import qualified Data.IntMap as IM

import Data.Maybe
import Data.IORef
import System.IO
import System.IO.Unsafe


data AnalysisMap = AnalysisMap {
    nextToken :: Int,
    tokenCountMap :: IM.IntMap Int,
    tokenFunctions :: IM.IntMap String
  }

changeTokenCountMap :: (IM.IntMap Int -> IM.IntMap Int) -> AnalysisMap -> AnalysisMap
changeTokenCountMap f the_map = the_map { tokenCountMap = f (tokenCountMap the_map) }

-- By making this a non-inlineable unsafe action we ensure that only a single
-- global map is shared between all the modules that reference the map.
{-# NOINLINE analysisMapRef #-}
analysisMapRef :: IORef AnalysisMap
analysisMapRef = unsafePerformIO $ newIORef (AnalysisMap 0 IM.empty IM.empty)

-- Creates a new token using the module/function name that we want to track with it
newToken :: String -> Int
newToken name = unsafePerformIO $ atomicModifyIORef analysisMapRef takeNextToken
  where
    takeNextToken analysis_map = let last_token = nextToken analysis_map
                                     next_token = last_token + 1
                                     token_functions' = IM.insert last_token name (tokenFunctions analysis_map)
                                 in (analysis_map { nextToken = next_token, tokenFunctions = token_functions' }, last_token)

-- Increments the call count associated with a token
incrementToken :: Int -> a -> a
incrementToken token before_what = unsafePerformIO (modifyIORef analysisMapRef (changeTokenCountMap (IM.insertWith (+) token 0))) `seq` before_what

-- Dump the accumulated statistics out to stderr
dumpStatistics :: IO ()
dumpStatistics = do
    analysis_map <- readIORef analysisMapRef
    let AnalysisMap { tokenCountMap = count_map, tokenFunctions = function_map } = analysis_map
        named_counts = IM.foldWithKey (\token count rest -> let function = fromJust $ IM.lookup token function_map
                                                            in (function ++ ": " ++ show count) : rest) [] count_map
    hPutStrLn stderr $ unlines named_counts

-- This is just here because it's tricky to look up (>>) at a specific type (we can
-- only really look up (>>) as a member of the Monad class)
thenIO :: IO a -> IO b -> IO b
thenIO = (>>)